/* ====================================================================== *\
   MANCALA.C -- Evaluate Mancala positions.

   Copyright (C) Jeff Erickson, 1994.  All rights reserved.
   Permission is given to copy this file as part of David Wolfe's
   games program, provided nobody makes any money in the process.
   ======================================================================
   This module calculates values of Mancala positions, as defined in
   the Klutz book.  The board consists of two rows of "pots", and two
   "stores", arranged as follows: 

			   /\ () () () () () () /\
			   \/ () () () () () () \/

   The upper pots and the left store belong the Left player; the
   others belong to the right player.  Legal moves are similar to
   those in Conway's Sowing game.  (Actually, Sowing moves are similar
   to those in Mancala!)  To move, a player pickes up0 all the stones
   in one of his pots and drops them one at a time into succeeding
   pots, always moving counterclockwise, including his own store, but
   NOT his opponents.  Normally, that's it, but there are two special
   types of moves:

   1. CAPTURE -- If the last stone lands in an empty pot on the
   players side, any stones in the opposite pot are revoved and put in
   the player's soter, along with the capturing stone.

   2. REPEAT -- If the last stone lands in the player's store, the
   player must move again.

   If one of the players cannot otherwise legally move, he must take
   all the stones in his opponent's pots and move them to his store.
   (Note that this might look like the null move.)  This ends the
   game.  The winner is the player with more stones in his store.  If
   multiple games are being played, the VALUE of the final position is
   half the difference between the number of stones in Left's store
   and the number in Right's store.

   In a typical opening position, each row has size pots, each pot
   has three stones, and the stores are empty.  This position is
   printed as follows: 

			    0 < 3 3 3 3 3 3 <
			      > 3 3 3 3 3 3 > 0

   ======================================================================
   Positions are stored in an array in the following order, where L is
   the "length" of the board.

		    L-1 < L-2 L-3 ...  1    0   <
			>  L  L+1 ... 2L-3 2L-2 > 2L-1

   Stones always move to higher-numbered slots (mod 2L) in the array.
\* ====================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "operations.h"
#include "output.h"
#include "mancala.h"

static char man_score = 1;
static long man_key = MANCALA_KEY;


/* ---------------------------------------------------------------------- *\
|  void man_print(man_type pos, int level)
|
|  print a mancala position at a given trace level
\* ---------------------------------------------------------------------- */
void man_print(man_type pos, int level)
{
    int i;

    for (i=0; i<level; i++)
	putchar('|'); 
    putchar(' ');

    printf("%2d <", pos->bd[pos->len-1]);
    for (i = pos->len-2; i>=0; i--)
	printf("%2d", pos->bd[i]);
    printf(" <\n");

    for (i=0; i<level; i++)
	putchar('|');
    putchar(' ');

    printf("   >");
    for (i = 0; i < pos->len-1; i++)
	printf("%2d", pos->bd[pos->len + i]);
    printf(" > %-2d", pos->bd[2*pos->len - 1]);
}

/* ---------------------------------------------------------------------- *\
|  void man_trace(man_type pos, game_type g, char *msg, int level)
|
|  Trace output.  Print a position, its value (if known), and an
|  optional message.  Do nothing unless TRACE_FLAG is set.  Inset the
|  output 'level' times.
\* ---------------------------------------------------------------------- */
void man_trace(man_type pos, game_type g, char *msg, int level)
{
    int i;
    game_type h;

    if (!(flags&TRACE_FLAG))
	return;
    if ((tlevel != 0) && (level > tlevel))
	return;

    if (pos)
	man_print(pos, level);
    else
    {
	for (i=0; i<level; i++)
	    putchar('|');
	putchar(' ');
    }

    if (g != NO_GAME)
    {
	printf(" = ");

	if (flags & ISH_FLAG)
	{
	    h = tweak(g);
	    if (eq(g, h))
		game_print(g);
	    else
	    {
		game_print(h);
		printf("-ish");
	    }
	}
	else
	    game_print(g);
    }

    if (msg)
	printf("        <%s>", msg);

    putchar('\n');
}


/* ---------------------------------------------------------------------- *\
|  man_type man_new(void)
|
|  Get a new mancala position
\* ---------------------------------------------------------------------- */
man_type man_new(void)
{
    man_type man;
    int i;

    man = (man_type) malloc(sizeof(struct _man_));
    man->len = MAX_MANCALA_LEN;
    man->score = 0;
    man->bd = (char *) malloc(sizeof(char) * MAX_MANCALA_LEN * 2);
    for (i=0; i < 2*man->len; i++)
	man->bd[i] = 0;
    return man;
}

/* ---------------------------------------------------------------------- *\
|  void man_copy(man_type new, man_type old)
|
|  Make a copy of a mancala position
\* ---------------------------------------------------------------------- */
void man_copy(man_type new, man_type old)
{
    int i;

    new->len = old->len;
    for (i=0; i<2*new->len; i++)
	new->bd[i] = old->bd[i];
    new->score = old->score;
}

/* ---------------------------------------------------------------------- *\
|  void man_free(man_type pos)
|
|  Free a mancala position
\* ---------------------------------------------------------------------- */
#define man_free(pos) if(pos){free(pos->bd);free(pos);}


/* ---------------------------------------------------------------------- *\
|  list_type man_to_list(man_type pos)
|
|  Convert a man_type to a list for hashing.  Try to use as short a
|  list as possible.
\* ---------------------------------------------------------------------- */
list_type man_to_list(man_type pos)
{
    list_type l;
    int i, base;
    int digit;

    l = list_make();
    list_insert(l, -pos->len);

    base = 2;
    for (i=0; i<2*pos->len ; i++)
	if (base < pos->bd[i]+2)
	    base = pos->bd[i]+2;
    list_insert(l, base*10000);

    digit = 1;
    for (i=0; i<2*pos->len; i++)
    {
	if (!man_score && ((i+1) % pos->len == 0))
	    continue;

	if (digit*base > (1L<<20))
	{
            list_insert(l, (i<<20) + digit);
	    digit = 1;
	}
	digit *= base;
	digit += pos->bd[i]+1;
    }
    list_insert(l, (i<<20) + digit);

#if 0
    man_print(pos);
    printf(" ==> ");
    list_printf(l);
#endif

    return l;
}


/* ---------------------------------------------------------------------- *\
|  man_type input_mancala(int length)
|
|  Ask the user to input a mancala position and return it.  If length
|  is zero, ask for length of board. 
\* ---------------------------------------------------------------------- */
man_type input_mancala(int length)
{
    man_type pos;
    char s[MAXBUFF];
    int i, j, space;

    pos = man_new();

    if (length == 0)
    {
	printf("Board length:\t");
	gets(s);
	pos->len = 0;
	sscanf(s, "%d", &pos->len);
	pos->len++;
	if (pos->len <= 1)
	{
	    printf("Huh?!\n");
	    man_free(pos);
	    return NULL;
	}
    }
    else
	pos->len = length+1;

    if (pos->len > MAX_MANCALA_LEN)
    {
	printf("Sorry, length can be at most %d\n", MAX_MANCALA_LEN-1);
	man_free(pos);
	return NULL;
    }

    printf("Top half:\t");
    for (i=0; i<pos->len; i++)
	printf("0 ");
    for (i=0; i<pos->len; i++)
	printf("\b\b");

    gets(s);
    space = 1;
    for (i=0, j=pos->len-1; s[i] && j>=0; i++)
    {
	if (isspace(s[i]))
	{
	    j -= !space;
	    space = 1;
	}
	else if (isdigit(s[i]))
	{
	    space = 0;
	    pos->bd[j] = pos->bd[j]*10 + s[i] - '0';
	}
	else
	{
	    printf("Unknown character '%c'\n", s[i]);
	    man_free(pos);
	    return NULL;
	}
    }

    printf("Bottom half:\t  ");
    for (i=0; i<pos->len; i++)
	printf("0 ");
    for (i=0; i<pos->len; i++)
	printf("\b\b");
    gets(s);
    space = 1;
    for (i=0, j=pos->len; s[i] && j<2*pos->len; i++)
    {
	if (isspace(s[i]))
	{
	    j += !space;
	    space = 1;
	}
	else if (isdigit(s[i]))
	{
	    space = 0;
	    pos->bd[j] = pos->bd[j]*10 + s[i] - '0';
	}
	else
	{
	    printf("Unknown character '%c'\n", s[i]);
	    man_free(pos);
	    return NULL;
	}
    }

    putchar('\n');
    return pos;
}


/* ---------------------------------------------------------------------- *\
|  void man_left_options(man_type pos, list_type options, int level)
|
|  Append a list of Left's options to the given list.
\* ---------------------------------------------------------------------- */
void man_left_options(man_type pos, list_type options, int level)
{
    int i, count, pot, moved;
    man_type new = man_new();

    moved = 0;
    for (i=0; i<pos->len-1; i++)
    {
	if (pos->bd[i])
	{
	    man_copy(new, pos);

	    count = new->bd[i];
	    new->bd[i] = 0;

	    pot = i;
	    while (count)
	    {
		pot++;
		if (pot == 2*pos->len-1)	/* skip opposing store */
		    pot = 0;

		new->bd[pot]++;
		count--;
	    }

	    if (pot == pos->len-1)		/* repeat */
	    {
		man_trace(new, NO_GAME, "Lrepeat", level+1);
		man_left_options(new, options, level+1);
	    }
	    else
	    {
		if ((pot < pos->len) && (new->bd[pot] == 1))	/* capture */
		{
		    new->bd[pot] = 0;
		    new->bd[pos->len-1] += new->bd[2*pos->len - pot - 2]+1;
		    new->bd[2*pos->len - pot - 2] = 0;
		}

		list_insert(options, mancala_position(new, level+1));
	    }
	    moved = 1;
	}
    }

    if (!moved && man_score)		/* game ends after repeat */
    {
	count = 0;
	for (i=0; i<pos->len; i++)
	    count += pos->bd[i];
	for (i=0; i<pos->len; i++)
	    count -= pos->bd[pos->len+i];

	man_trace(pos, num(Q_int(count)), "Lend", level);
	list_insert(options, num(Q_int(count)));
    }

    man_free(new);
}

/* ---------------------------------------------------------------------- *\
|  void man_right_options(man_type pos, list_type options, int level)
|
|  Append a list of Right's options to the given list.  Looks
|  familiar, doesn't it?
\* ---------------------------------------------------------------------- */
void man_right_options(man_type pos, list_type options, int level)
{
    int i, count, pot, moved;
    man_type new = man_new();

    moved = 0;
    for (i=pos->len; i<(2*pos->len - 1); i++)
    {
	if (pos->bd[i])
	{
	    man_copy(new, pos);

	    count = new->bd[i];
	    new->bd[i] = 0;

	    pot = i;
	    while (count)
	    {
		pot = (pot+1)%(2*pos->len);
		if (pot == pos->len - 1)	/* skip opposing store */
		    pot++;

		new->bd[pot]++;
		count--;
	    }

	    if (pot == 2*pos->len - 1)		/* repeat */
	    {
		man_trace(new, NO_GAME, "Rrepeat", level+1);
		man_right_options(new, options, level+1);
	    }
	    else
	    {
		if ((pot >= pos->len) && (new->bd[pot] == 1))	/* capture */
		{
		    new->bd[pot] = 0;
		    new->bd[2*pos->len-1] += new->bd[2*pos->len - pot - 2]+1;
		    new->bd[2*pos->len - pot - 2] = 0;
		}
		list_insert(options, mancala_position(new, level+1));
	    }
	    moved = 1;
	}
    }

    if (!moved && man_score)		/* game ends after repeat */
    {
	count = 0;
	for (i=0; i<pos->len; i++)
	    count += pos->bd[i];
	for (i=0; i<pos->len; i++)
	    count -= pos->bd[pos->len+i];

	man_trace(pos, num(Q_int(count)), "Rend", level);
	list_insert(options, num(Q_int(count)));
    }

    man_free(new);
}

/* ---------------------------------------------------------------------- *\
|  game_type mancala_position(man_type pos, int level)  
|
|  Evaluate a Mancala position, and memoize the result.
\* ---------------------------------------------------------------------- */
game_type mancala_position(man_type pos, int level)
{
    int i, count;
    int oldl, oldr;
    list_type l;
    list_type left, right;
    game_type g;

    oldl = pos->bd[pos->len-1];
    oldr = pos->bd[2*pos->len-1];
    pos->bd[pos->len-1] = 0;
    pos->bd[2*pos->len-1] = 0;

    count = 0;
    for (i=0; i<pos->len-1; i++)
    {
	count += pos->bd[i];
	count += pos->bd[pos->len + i];
    }

    if (!count)
    {
	g = zero;
	man_trace(pos, g, "trivial", level);
    }
    else
    {
	l = man_to_list(pos);
	if (hash_test(man_key, l))
	{
	    list_free(l);
	    g = hash_get_last();
	    man_trace(pos, g, "seen it", level);
	}
	else
	{
	    man_trace(pos, NO_GAME, NULL, level);

	    left = list_make();
	    man_left_options(pos, left, level);
	    right = list_make();
	    man_right_options(pos, right, level);

	    g = make(left, right);
	    man_trace(pos, g, NULL, level);
		 
	    hash_put(man_key, l, g); 	/* memoize, ignoring score */
	}
    }

    if (man_score && (oldl || oldr))
    {
	pos->bd[pos->len-1] = oldl;
	pos->bd[2*pos->len-1] = oldr;
	g = plus(g, num(Q_int(oldl - oldr)));
	man_trace(pos, g, "+ score", level);
    }

    return g;
}


/* ---------------------------------------------------------------------- *\
|  game_type mancala(int max, int length)
|
|  Get a Mancala position and evaluate it.  If called with nonzero
|  arguments, enumerate all positions with the given length and at
|  most max stones in each pot.
\* ---------------------------------------------------------------------- */
game_type mancala(int max, int length)
{
    game_type g;
    man_type pos;
    int count,i,j;

    man_key = MANCALA_KEY;
    man_score = 1;

    if (max && length)
    {
	pos = man_new();
	pos->len = ++length;

	while (1)
	{
	    for (i=0; i<2*length-1; i++)
	    {
		if (i == length-1)
		    continue;
		if (pos->bd[i] != max)
		    break;
	    }

	    if (i == 2*length-1)
		break;

	    for (j=0; j<i; j++)
		pos->bd[j]= 0;
	    pos->bd[i]++;

	    g = mancala_position(pos, 0);

	    man_print(pos, 0);
	    printf(" = ");
	    game_printf(g);
	}

	return NO_GAME;
    }

    pos = input_mancala(0);
    if (pos == NULL)
	return NO_GAME;

    g = mancala_position(pos, 0);

    man_free(pos);

    return g;
}

/* ---------------------------------------------------------------------- *\
|  game_type revmancala(int left, int right)
|
|  Get a Reverse Mancala position and evaluate it, whatever that means.
\* ---------------------------------------------------------------------- */
game_type revmancala(int left, int right)
{
    printf("Huh?\n");
    return NO_GAME;
}

/* ---------------------------------------------------------------------- *\
|  game_type hotmancala(int left, int right)
|
|  Get a Hot Mancala position and evaluate it.
\* ---------------------------------------------------------------------- */
game_type hotmancala(int left, int right)
{
    printf("Huh?\n");
    return NO_GAME;
}

/* ---------------------------------------------------------------------- *\
|  game_type coldmancala(int max, int length)
|
|  Get a Cold Mancala position and evaluate it.
\* ---------------------------------------------------------------------- */
game_type coldmancala(int max, int length)
{
    game_type g;
    man_type pos;
    int count,i,j;

    man_key = COLDMANCALA_KEY;
    man_score = 0;

    if (max && length)
    {
	pos = man_new();
	pos->len = ++length;

	while (1)
	{
	    for (i=0; i<2*length-1; i++)
	    {
		if (i == length-1)
		    continue;
		if (pos->bd[i] != max)
		    break;
	    }

	    if (i == 2*length-1)
		break;

	    for (j=0; j<i; j++)
		pos->bd[j]= 0;
	    pos->bd[i]++;

	    g = mancala_position(pos, 0);

	    man_print(pos, 0);
	    printf(" = ");
	    game_printf(g);
	}

	return NO_GAME;
    }

    pos = input_mancala(0);
    if (pos == NULL)
	return NO_GAME;

    g = mancala_position(pos, 0);

    man_free(pos);

    return g;
}

