/* ====================================================================== *\
   SOWING.C -- Evaluate Conway's Sowing positions.

   Copyright (C) Jeff Erickson, 1994.  All rights reserved.
   Permission is given to copy this file as part of David Wolfe's
   games program, provided nobody makes any money in the process.
   ======================================================================
   This module calculates values of Sowing positions.  To the author's
   knowledge, Sowing was first described by John Conway at the
   Combinatorial Games Workshop at MSRI, July 1994.  The game is
   played with a set of beans and a line of pots.  A legal move
   consists of picking up all the beans in any pot, and moving towards
   the opposing player, dropping one bean in each pot, subject to the
   restriction that the last bean must not land in an empty pot.  The
   player who moves last wins.

   For example, the position 1223 is evaluated as follows:

	   1223 = {323, 1034 | 2303, 2330}
	        = {0, 0 | { | 341}, {44 | }}
		= {0 | { | { | 35}}, 1}
		= {0 | { | { | 0}}}
		= 0|-2

   Also plays Sowing impartially:

	   1223 = {323, 1034, 2303, 2330}
	        = {0, 0, {341}, {44}}
		= {0, {{35}}, {0}}
		= {0, {{0}}, *}
		= *2

   If the code is invoked with nonzero arguments, all sowing
   positions, with up to the specified size and number of stones, are
   evaluated and printed.  The enumeration code attempts to detect
   positions that can be split into independent components.  Splitting
   positions in the general evaluation code would almost certainly not
   be worthwhile, since splits are relatively rare.
\* ====================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "operations.h"
#include "output.h"
#include "sowing.h"


/* ---------------------------------------------------------------------- *\
|  sow_type sow_new(int length)
|
|  Get a new sowing position with <length> slots
\* ---------------------------------------------------------------------- */
sow_type sow_new(int length)
{
    sow_type sow;
    int i;

    sow = (sow_type) malloc(sizeof(struct sow));
    sow->length = length;
    sow->first = 1;
    sow->last = length;
    sow->board = (char *) malloc((length+2) * sizeof(char));
    for (i=0; i<length+2; i++)
	sow->board[i] = 0;
    return sow;
}

/* ---------------------------------------------------------------------- *\
|  void sow_free(sow_type pos)
|
|  Free a sowing position
\* ---------------------------------------------------------------------- */
#define sow_free(pos) if(pos){free((pos)->board);free(pos);}


/* ---------------------------------------------------------------------- *\
|  void sow_print(sow_type pos)
|
|  print a sowing position
\* ---------------------------------------------------------------------- */
void sow_print(sow_type pos)
{
    int i, beans;

    for (i = pos->first; i <= pos->last; i++)
    {
	beans = pos->board[i];
	if (beans < 0)
	    putchar('x');
	else if ((beans > i - pos->first) && (beans > pos->last - i))
	    putchar('X');
	else if (beans < 10)
	    putchar(beans + '0');
	else
	    putchar(beans - 10 + 'a');
    }
}

/* ---------------------------------------------------------------------- *\
|  void sow_trace(sow_type pos, game_type g, int s, char *msg, int level)
|
|  Trace output.  Print a position, its value (if known), and an
|  optional message.  Do nothing unless TRACE_FLAG is set.  Inset the
|  output 'level' times.
\* ---------------------------------------------------------------------- */
void sow_trace(sow_type pos, game_type g, int s, char *msg, int level)
{
    int i;

    if (!(flags&TRACE_FLAG))
    {
	if (!s && !msg && (flags & ISH_FLAG))	/* not the best choice */
	{
	    sow_print(pos);
	    printf(" = 0\n");
	}
	return;
    }
    if ((tlevel != 0) && (level > tlevel))
	return;

    for (i=0; i<level; i++)
	putchar('|'); 
    putchar(' ');

    sow_print(pos);

    if (g != NO_GAME)
    {
	printf(" = ");
	game_print(g);
    }
    else if (s != -1)
    {
	if (s == 0)
	    printf(" = 0");
	else if (s == 1)
	    printf(" = *");
	else
	    printf(" = *%d", s);
    }

    if (msg)
	printf("        <%s>", msg);

    putchar('\n');
}


/* ---------------------------------------------------------------------- *\
|  void sow_simplify(sow_type pos)
|
|  Simplify a position (in place).
\* ---------------------------------------------------------------------- */
void sow_simplify(sow_type pos)
{
    int i, beans;

    for (i=1; (i<=pos->length) && (pos->board[i] == 0); i++)
	/* do nothing */;
    pos->first = i;

    for (i=pos->length ; (i>=pos->first) && (pos->board[i] == 0); i--);
	/* do nothing */;
    pos->last = i;

    for (i=pos->first; i<=pos->last; i++)
    {
	beans = pos->board[i];
	if ((beans > i - pos->first) && (beans > pos->last - i))
	    pos->board[i] = -1;
    }
}

/* ---------------------------------------------------------------------- *\
|  sow_type input_sowing(void)
|
|  Ask the user to input a sowing position and return it.
\* ---------------------------------------------------------------------- */
sow_type input_sowing(void)
{
    sow_type pos;
    char s[MAXBUFF];
    int i,j;
    char space=1;
    
    printf("Enter position (list of hexidecimal digits):\n");
    gets(s);

    pos = sow_new(strlen(s));

    for (i=0, j=1; s[i]; i++)
    {
	if (isspace(s[i]))
	    continue;
	else if (isdigit(s[i]))
	    pos->board[j++] = s[i] - '0';
	else if (isalpha(s[i]) && (tolower(s[i]) <= 'f'))
	    pos->board[j++] = tolower(s[i]) - 'a' + 10;
	else if (tolower(s[i]) == 'x')
	    pos->board[j++] = -1;
	else
	{
	    printf("Unknown character '%c'\n", s[i]);
	    sow_free(pos);
	    return NULL;
	}
    }
    pos->length = j;
    sow_simplify(pos);
    return pos;
}


/* ---------------------------------------------------------------------- *\
|  list_type sow_to_list(sow_type pos)
|
|  Convert a sow_type to a list.  [Wolfe should have called them "sets"
|  instead of "lists", since insertion order doesn't matter!]  Saves
|  space by combining small integers into larger ones.  What this
|  program really needs is the ability to has STRINGS.
\* ---------------------------------------------------------------------- */
list_type sow_to_list(sow_type pos)
{
    list_type l;
    int i, base, digit;

    l = list_make();
    list_insert(l, pos->last - pos->first);

    base = 2;
    for (i=pos->first; i<=pos->last; i++)
	if (base < pos->board[i] + 2)
	    base = pos->board[i] + 2;
    list_insert(l, base*10000);

    digit = 0;
    for (i = 0; i <= pos->last - pos->first; i++)
    {
	if (digit*base > (1L<<20))
	{
	    list_insert(l, (i<<20) + digit);
	    digit = 0;
	}
	digit *= base;
	
	digit += pos->board[pos->first + i] + 1;
    }
    if (digit != 0)
	list_insert(l, (i<<20) + digit);

#ifdef EBUG
    sow_print(pos);
    printf("  ==>  ");
    list_printf(l);
#endif

    return l;
}

/* ---------------------------------------------------------------------- *\
|  list_type symmetric_sow_to_list(sow_type pos)
|
|  Convert a sow_type to a list.  Compares the position with its
|  mirror image and build the list from the lexicographically smaller
|  of the pair.  Saves space by combining small integers into larger
|  ones.  What this program really needs is a good five-cent cigar.
\* ---------------------------------------------------------------------- */
list_type symmetric_sow_to_list(sow_type pos)
{
    list_type l;
    int i, base, sign;
    int beans, snaeb, digit;

    l = list_make();
    list_insert(l, pos->last - pos->first);

    for (sign=i=0; !sign && (pos->first+i <= pos->last-i); i++)
    {
	beans = pos->board[pos->first + i];
	if (beans < 0)
	    beans = 10000;

	snaeb = pos->board[pos->last - i];
	if (snaeb < 0)
	    snaeb = 10000;

	sign = snaeb - beans;
    }

    base = 2;
    for (i=pos->first; i<=pos->last; i++)
	if (base < pos->board[i] + 2)
	    base = pos->board[i] + 2;
    list_insert(l, base*10000);

    if (sign < 0)
    {
	digit = 0;
	for (i = 0; i <= pos->last - pos->first; i++)
	{
	    if (digit*base > (1L<<20))
	    {
		list_insert(l, (i<<20) + digit);
		digit = 0;
	    }
	    digit *= base;
	    digit += pos->board[pos->last - i] + 1;
	}
	if (digit != 0)
	    list_insert(l, (i<<20) + digit);
    }
    else
    {
	digit = 0;
	for (i = 0; i <= pos->last - pos->first; i++)
	{
	    if (digit*base > (1L<<20))
	    {
		list_insert(l, (i<<20) + digit);
		digit = 0;
	    }
	    digit *= base;

	    digit += pos->board[pos->first + i] + 1;
	}
	if (digit != 0)
	    list_insert(l, (i<<20) + digit);
    }

#ifdef EBUG
    putchar(sign<0 ? '-' : '+');
    sow_print(pos);
    printf(" (%d to %d) ==>  ", pos->first, pos->last);
    list_printf(l);
#endif

    return l;
}


/* ---------------------------------------------------------------------- *\
|  int mex(list_type list)
|
|  Compute the mex of a list of integers.  Used for evaluating
|  impartial games more quickly than with the generic evaluation code.
\* ---------------------------------------------------------------------- */
int mex(list_type list)
{
    int m = 0;

    list = list->next;		/* skip header node */
    while (list != NULL)
    {
	if (list->elt > m)
	    return m;
	else if (list->elt == m)
	    m++;
	list = list->next;
    }
    return m;
}

/* ---------------------------------------------------------------------- *\
|  int isowing_position(sow_type pos, int level)
|
|  Evaluate a impartial sowing position, and memoize the result.
|  NOTE THAT THIS FUNCTION RETURNS AN INTEGER, NOT A GAME_TYPE!!
\* ---------------------------------------------------------------------- */
int isowing_position(sow_type pos, int level)
{
    register int i, j, beans;
    list_type options;
    int g, h;		/* really games */
    list_type l;
    sow_type new;

    /* ---- Check for trivial cases ---- */

    if (pos->last - pos->first < 1)
	return 0;

    /* ---- Copy into canonical form ---- */

    new = sow_new(pos->last - pos->first + 1);
    for (i=0; i <= pos->last - pos->first; i++)
    {
	beans = pos->board[pos->first+i];
	if ((beans > i) && (beans > pos->last - pos->first - i))
	    beans = -1;
	new->board[i+1] = beans;
    }

    /* ---- Check for earlier value ---- */
    
    l = symmetric_sow_to_list(new);
    if (hash_test(ISOWING_KEY, l))
    {
	list_free(l);
	g = hash_get_last();
	sow_trace(new, NO_GAME, g, "seen it", level);
	return g;
    }

    /* ---- Recusively evaluate moves ---- */

    sow_trace(new, NO_GAME, -1, NULL, level);

    options = list_make();		/* options */

    for (i=new->first; i<=new->last; i++)
    {
	beans = new->board[i];
	if (beans <= 0)
	    continue;

	if ((beans+i <= new->last) && (new->board[i+beans] != 0))
	{
	    new->board[i] = 0;
	    for (j=1; j<=beans; j++)
		if (new->board[i+j] >= 0)
		    new->board[i+j]++;
	    if (i == new->first)
		new->first++;

	    list_insert(options, isowing_position(new, level+1));

	    if (i < new->first)
		new->first--;
	    new->board[i] = beans;
	    for (j=1; j<=beans; j++)
		if (new->board[i+j] > 0)
		    new->board[i+j]--;
	}

	if ((i-beans >= new->first) && (new->board[i-beans] != 0))
	{
	    new->board[i] = 0;
	    for (j=1; j<=beans; j++)
		if (new->board[i-j] >= 0)
		    new->board[i-j]++;
	    if (i == new->last)
		new->last--;

	    list_insert(options, isowing_position(new, level+1));

	    if (i > new->last)
		new->last++;
	    new->board[i] = beans;
	    for (j=1; j<=beans; j++)
		if (new->board[i-j] > 0)
		    new->board[i-j]--;
	}
    }

    g = mex(options);
    list_free(options);

    hash_put(ISOWING_KEY, l, g);		/* memoize */

    sow_trace(new, NO_GAME, g, NULL, level);
    sow_free(new);
    return g;
}


/* ---------------------------------------------------------------------- *\
|  game_type sowing_position(sow_type pos, int level)
|
|  Evaluate a partizan sowing position, and memoize the result.
\* ---------------------------------------------------------------------- */
game_type sowing_position(sow_type pos, int level)
{
    register int i, j, beans;
    list_type left, right;
    game_type g, h;
    sow_type new;
    list_type l;

    /* ---- Check for trivial cases ---- */

    if (pos->last - pos->first < 1)
	return zero;

    /* ---- Copy into canonical form ---- */

    new = sow_new(pos->last - pos->first + 1);
    for (i=0; i <= pos->last - pos->first; i++)
    {
	beans = pos->board[pos->first+i];
	if ((beans > i) && (beans > pos->last - pos->first - i))
	    beans = -1;
	new->board[i+1] = beans;
    }

    /* ---- Check for earlier value ---- */
    
    l = sow_to_list(new);
    if (hash_test(SOWING_KEY, l))
    {
	list_free(l);
	g = hash_get_last();
	sow_trace(new, g, -1, "seen it", level);
	return g;
    }

    /* ---- Recusively evaluate moves ---- */

    sow_trace(new, NO_GAME, -1, NULL, level);

    left = list_make();			/* left options */
    right = list_make();		/* right options */

    for (i=new->first; i<=new->last; i++)
    {
	beans = new->board[i];
	if (beans <= 0)
	    continue;

	if ((beans+i <= new->last) && (new->board[i+beans] != 0))
	{
	    new->board[i] = 0;
	    for (j=1; j<=beans; j++)
		if (new->board[i+j] >= 0)
		    new->board[i+j]++;
	    if (i == new->first)
		new->first++;

	    list_insert(left, sowing_position(new, level+1));

	    if (i < new->first)
		new->first--;
	    new->board[i] = beans;
	    for (j=1; j<=beans; j++)
		if (new->board[i+j] > 0)
		    new->board[i+j]--;
	}

	if ((i-beans >= new->first) && (new->board[i-beans] != 0))
	{
	    new->board[i] = 0;
	    for (j=1; j<=beans; j++)
		if (new->board[i-j] >= 0)
		    new->board[i-j]++;
	    if (i == new->last)
		new->last--;

	    list_insert(right, sowing_position(new, level+1));

	    if (i > new->last)
		new->last++;
	    new->board[i] = beans;
	    for (j=1; j<=beans; j++)
		if (new->board[i-j] > 0)
		    new->board[i-j]--;
	}
    }

    g = make(left,right);

    hash_put(SOWING_KEY, l, g);		/* memoize */

    sow_trace(new, g, -1, NULL, level);
    sow_free(new);
    return g;
}


/* ---------------------------------------------------------------------- *\
|  int sowing_enumerate(int impartial, int length, int beans)
|
|  Enumerate all sowing positions with the sepcified number of pots
|  and beans.  Evaluate each position only once, and only if it is
|  nontrivial.  Try to detect positions that can be split into
|  multiple components.  Return the number of positions evaluated.
\* ---------------------------------------------------------------------- */
int sowing_enumerate(int impartial, int length, int beans)
{
    sow_type pos;
    game_type g;
    int i, j, val;
    int okay, count, numpos;

    for (count=i=0; i<length; i++)
	count += (i>length-i+1) ? i : (length-i+1);
    if (count <= beans)
	return 0;

    numpos = 0;
    printf("%d pots\t%d beans\n", length, beans);
    
    pos = sow_new(length);
    pos->board[1] = beans-1;
    for (i=2; i<length; i++)
	pos->board[i] = 0;
    pos->board[length] = 1;
    
    while (1)
    {
	/* ---- make sure there's at least one legal move ---- */

	okay = 0;
	for (i=1; !okay && i<=length; i++)
	{
	    count = pos->board[i];
	    if (count == 0)
		continue;
	    else if ((count < i) && (pos->board[i-count] > 0))
		okay = 1;
	    else if ((count <= length-i) && (pos->board[i+count] > 0))
		okay = 1;
	}

	/* ---- ignore positions with over-full pots ---- */

	for (i=1; okay && i<=length; i++)
	    if ((pos->board[i] > i) && (pos->board[i] > length-i+1))
		okay = 0;

	/* ---- look for mirror images ---- */

	for (i=1; okay && i <= (length+1)/2; i++)
	{
	    if (pos->board[i] > pos->board[length+1-i])
		break;
	    else if (pos->board[i] < pos->board[length+1-i])
		okay = 0;
	}

	/* ---- look for EASILY splittable positions ---- */

	/* ---- if it's a new nontrivial position, evaluate it ---- */

	if (okay)
	{
	    if (impartial)
	    {
		val = isowing_position(pos, 0);
		g = star(val);
	    }
	    else
	    {
		val = 100;
		g = sowing_position(pos, 0);
	    }

	    if (val > -1)
	    {
		printf("\t\t");
		sow_print(pos);
		printf(" = ");
		game_printf(g);
	    }

	    numpos++;
	}

	/* ---- make next position ---- */

	pos->board[1]--;
	pos->board[length]--;

	for (i=1; i<length; i++)
	    if (pos->board[i] != 0)
		break;

	if (i == length)	/* All done! */
	    break;

	pos->board[i]--;
	pos->board[i+1]++;
	if (i>1)
	{
	    pos->board[1] = pos->board[i];
	    pos->board[i] = 0;
	}

	pos->board[1]++;
	pos->board[length]++;
    }

    return numpos;
}

/* ---------------------------------------------------------------------- *\
|  void sowing_enumerate_to(int impartial, int maxlength, int maxbeans)
|
|  Enumerate all sowing positions with up to maxlength pots and
|  maxbeans beans.  Evaluate each position only once, and only if it
|  is nontrivial.  Try to detect positions that can be split into
|  multiple components.
\* ---------------------------------------------------------------------- */
void sowing_enumerate_to(int impartial, int maxlength, int maxbeans)
{
    int length, beans, count;
	
    printf("Positions with at most %d pots and at most %d beans\n",
	   maxlength, maxbeans);
    printf("---------------------------------------------------\n");

    for (length = 2; length <= maxlength; length++)
	for (beans = (length+3)/2; beans <= maxbeans; beans++)
	   if (!sowing_enumerate(impartial, length, beans))
	       break;
}


/* ---------------------------------------------------------------------- *\
|  game_type sowing(int impartial, int length, int beans)
|
|  Get a Sowing position and evaluate it.
\* ---------------------------------------------------------------------- */
game_type sowing(int impartial, int length, int beans)
{
    game_type g;
    sow_type pos;

    if (!length && !beans)
    {
	pos = input_sowing();
	if (pos == NULL)
	    return NO_GAME;

	if (impartial)
	    g = star(isowing_position(pos, 0));
	else
	    g = sowing_position(pos, 0);

	sow_free(pos);
	return g;
    }
    else
    {
	if (length == 0)
	{
	    for (length = 2; beans >= (length+3)/2; length++)
		sowing_enumerate(impartial, length, beans);
	}
	else if (beans == 0)
	{
	    for (beans = (length+3)/2; ; beans++)
		if (!sowing_enumerate(impartial, length, beans))
		    break;
	}
	else if (beans >= (length+3)/2)
	{
	    if (!sowing_enumerate(impartial, length, beans))
		printf("Too many beans!  All those positions are trivial!\n");
	}
	else
	    printf("Too few beans!  All those positions are splittable!\n");
	return NO_GAME;
    }
}
