/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <stdlib.h>
#include "externs.h"
#include "list.h"
#include "bitmatrix.h"

B_type B_make (int x, int y)
{
	B_type a;
	int i;

	if (!(a = (B_type) malloc (sizeof (struct B_struct))))
		myerror ("B_make: malloc couldn't free block");
	a->x = x;
	a->y = y;
	if (!(a->contents = (unsigned *) malloc (((x*y+(INT_SIZE - 1)) / INT_SIZE) * sizeof (unsigned))))
		myerror ("B_make: malloc couldn't free block");
	for (i = 0; i < (x*y+(INT_SIZE - 1))/INT_SIZE; i++) *(a->contents + i) = (unsigned) 0;
	return a;
}

list_type B_to_list (B_type a)
{
	list_type l;
	int i;
	
	l = list_make();
	list_insert_unsorted (l, a->x);
	list_insert_unsorted (l, a->y);
	for (i = 0; i <= (a->x * a->y) / INT_SIZE; i++) list_insert_unsorted (l, *(a->contents + i));
	return l;
}

B_type B_pack (B_type a)
{
	B_type b;
	int i, j, xmin, ymin, xmax, ymax;

	xmin=a->x, ymin=a->y, xmax=-1, ymax=-1;
	for (i=0; i < a->x; i++)
		for (j=0; j < a->y; j++)
			if (B_get (a,i,j)) {
				if (i < xmin) xmin = i;
				if (j < ymin) ymin = j;
				if (i > xmax) xmax = i;
				if (j > ymax) ymax = j;
			}
	if (xmin == 0 && ymin==0 && xmax==a->x-1 && ymax==a->y-1) return a;
	if (xmin > xmax) {
		B_free (a);
		return B_make (0,0);
	}
	b = B_make (xmax - xmin + 1, ymax - ymin + 1);
	for (i = 0; i < xmax - xmin + 1; i++)
		for (j = 0; j < ymax - ymin + 1; j++)
			if (B_get (a, i+xmin, j+ymin)) B_set (b,i,j);
	B_free (a);
	return b;
}

void B_printf (B_type a) {
	int x,y;
	for (x = 0; x < a->x; x++) {
		for (y = 0; y < a->y; y++)
			printf ("%d", B_get (a,x,y) ? 1 : 0);
		printf ("\n");
	}
}

B_type B_copy (B_type a)
{
  B_type b;
  int i, j;

  b = B_make(B_x(a), B_y(a));

  for (i=0; i<B_x(a); i++)
    for (j=0; j<B_y(a); j++)
      if (B_get (a,i,j))
	B_set (b,i,j);
  return b;
}


B_type B_copy_offset (B_type old, int xoffset, int yoffset, int xsize, int ysize)
{
  B_type new;
  int i, j, k, l;

  new = B_make(xsize, ysize);

  for (i=0, k=xoffset; i<xsize; i++, k++)
    for (j=0, l=yoffset; j<ysize; j++, l++)
      if (B_get(old, k, l))
	B_set(new, i, j);
  return new;
}


/* Does not modify or free its argument; just returns an expanded copy. */
B_type B_expand (B_type a, int dx, int dy)
{
  B_type b;
  int i, j;

  /*
  if (KDEBUG && EXPAND_DEBUG)
    { kshowdepth();
      printf("Pre-expanded (%d,%d):\n",dx,dy);
      B_printf(a); }
      */

  b = B_make (B_x(a) + abs(dx), B_y(a) + abs(dy));

  /*
  if (KDEBUG && EXPAND_DEBUG)
    { kshowdepth();
      printf("New matrix (%d x %d):\n",B_x(b),B_y(b));
      B_printf(b); }
      */

  if (dx + dy > 0)
    { for (i=0; i < B_x(a); i++)
      for (j=0; j < B_y(a); j++)
	if (B_get (a,i,j))
	  B_set (b,i,j); }
  else
    /* dx or dy is -1, the other is 0 */
    { for (i=0; i < B_x(a); i++)
      for (j=0; j < B_y(a); j++)
	  if (B_get (a,i,j))
	    B_set (b,i-dx,j-dy); }
  return b;
}

/* frees its argument */
B_type B_transpose (B_type a)
{
  B_type b;
  int i, j;

  b = B_make(B_y(a), B_x(a));

  for (i=0; i<B_x(a); i++)
    for (j=0; j<B_y(a); j++)
      if (B_get (a,i,j))
	B_set (b,j,i);
  /* Does this freeing of a present a problem?  (I mean in my code.)
     I don't think so. */
  B_free(a);
  return b;
}



/* More orientation cleverness for efficient hashing:  If both of the
bottom corners are occupied, make sure that both of the top ones are too.
If only one of the top corners is occupied, make sure that the left-hand
one is.  */

/* I ignore edges because I started out doing so and it's simplest to
continue doing so.  I have no clear idea whether orienting according to
which edges are present would be better than this strategy, and no
particular reason to suspect it would be significantly so.  Also because
this is an operation on bitmatrices, not Konane representations. */

/* Returns an integer:  1s bit is set if it was flipped horizontally.
                        2s bit is set if it was flipped vertically. */
int B_flip (B_type a)
{ boolean temp;
  int i, j, k;
  boolean result = 0;
  
  /* perhaps consolidate these two flips */
  if (B_get(a,0,B_y(a)-1) & !B_get(a,0,0))
    /* flip horizontally */
    { for (i=0; i<B_x(a); i++)
	for (j=0,k=B_y(a)-1; j<k; j++,k--)
	  { temp = B_get(a,i,j);
	    if (B_get (a,i,k))
	      B_set (a,i,j);
	    else
	      B_clear (a,i,j);
	    if (temp)
	      B_set (a,i,k);
	    else
	      B_clear (a,i,k); }
      result = result+1; }
  if (B_get(a,B_x(a)-1,0) & B_get(a,B_x(a)-1,B_y(a)-1)
      & !(B_get(a,0,0) & B_get(a,0,B_y(a)-1)))
    /* flip vertically */
    { for (j=0; j<B_y(a); j++)
	for (i=0,k=B_x(a)-1; i<k; i++,k--)
	  { temp = B_get(a,i,j);
	    if (B_get (a,k,j))
	      B_set (a,i,j);
	    else
	      B_clear (a,i,j);
	    if (temp)
	      B_set (a,k,j);
	    else
	      B_clear (a,k,j); }
      result = result+2; }
  return result;
}


/* Returns number of first row with a stone, or -1 if no row has a stone. */
int B_min_occupied_x (B_type a)
{
  int i, j, xlimit, ylimit;

  xlimit = B_x(a);
  ylimit = B_y(a);

  for (i=0; i < xlimit; i++)
    for (j=0; j < ylimit; j++)
      if (B_get (a,i,j))
	return i;

  return -1;
}

/* Returns number of last row with a stone, or -1 if no row has a stone. */
int B_max_occupied_x (B_type a)
{
  int i, j, ylimit;

  ylimit = B_y(a);

  for (i=B_x(a)-1; i >= 0; i--)
    for (j=0; j < ylimit; j++)
      if (B_get (a,i,j))
	return i;

  return -1;
}


/* Returns number of first column with a stone, or -1 if no column has a stone. */
int B_min_occupied_y (B_type a)
{
  int i, j, xlimit, ylimit;

  xlimit = B_x(a);
  ylimit = B_y(a);


  for (j=0; j < ylimit; j++)
    for (i=0; i < xlimit; i++)
      if (B_get (a,i,j))
	return j;

  return -1;
}


/* Returns number of last column with a stone, or -1 if no column has a stone. */
int B_max_occupied_y (B_type a)
{
  int i, j, xlimit;

  xlimit = B_x(a);


  for (j=B_y(a)-1; j >= 0; j++)
    for (i=0; i < xlimit; i++)
      if (B_get (a,i,j))
	return j;

  return -1;
}
