/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* if (flags & ASSUME_FLAG) is true,
 * We assume that 1) players want to invade/block ONLY the longest
 * blocked or unblocked corridor of a given group.  2) White's
 * invasion of an unblocked corridor is REVERSED throught the blck of
 * that same corridor. */

#include <stdlib.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "operations.h"
#include "corridors.h"
#include "output.h"

list_type invasion [MAX_INVASIONS];
#ifdef IBMPC
unsigned invasions=0;
#else
unsigned long invasions=0;
#endif
int change_sizes = 1; /* "1/0" means assume size of invading groups "do/don't" matter */

extern void corridor_Print();

void corridor_clear(void) {
	unsigned long i;
	
	for (i = 1; i <= invasions; i++)
		list_free (invasion[i]);
	invasions = 0;
}

game_type blocked (int n) {
	return plus (g_int (n-2), mult (num (Q_exp (1-n)), one_star));
}

game_type unblocked (int n) {
	if (n == 0) return zero;
	else if (n == 1) return star (1);
	else return plus (g_int (n-4), mult (num (Q_exp (3-n)), one_star));
}

int corridor_get (list_type l) {
	if (hash_test (CORRIDOR_KEY, l)) return hash_get_last ();
	if (++invasions >= MAX_INVASIONS) {
		myerror ("corridor_get: No more invasions.  Sorry.");
		exit (1);
	}
	invasion[invasions] = list_copy (l);
	hash_put (CORRIDOR_KEY, list_copy (l), invasions);
	return invasions;
}

static int corridor_size (int i) {
	list_type l;

	for (l = list_start (invasion[i]); l != NIL; l = list_rest (l))
		if ((FLAGS & list_pick (l)) == SIZE)
			return (list_pick (l) & DATA);
	myerror ("corridor_size: No Size");
	return 0;
}

static int corridor_change (int stack[], int o1, int o2, int n1, int n2) {
	list_type inv;
	int i,j;

	inv = invasion[i = stack[stack[0]]];
	if (o1 != 0) list_delete (inv, o1);
	if (o2 != 0) list_delete (inv, o2);
	if (n1 != 0) list_insert (inv, n1);
	if (n2 != 0) list_insert (inv, n2);
	j = corridor_get (inv);
	if (n1 != 0) list_delete (inv, n1);
	if (n2 != 0) list_delete (inv, n2);
	if (o1 != 0) list_insert (inv, o1);
	if (o2 != 0) list_insert (inv, o2);
	if (stack[0] > 1) {
		stack[0]--;
		j = corridor_change (stack, FOLLOWER|i, 0, FOLLOWER|j, 0);
		stack[0]++;
	}
	return j;
}

/* An invasion is a list encoding size, parent, blocked, unblocked, follower */
/* Function used to construct left and right followers */
static void corridor_eval (int i, int stack[], list_type left, list_type right) {
	int size, parent=0, long_b=0, long_u=0, followers=0;
	list_type inv, l, l1, l2;
	int n, flgs, data, t;
	game_type g;

	if (stack[0] != 1) parent = stack[stack[0]-1];
	inv = list_copy (invasion [i]);
	for (l = list_start (inv); l != NIL; l = list_rest (l)) {
		n = list_pick (l);
		flgs = n & FLAGS;
		data = n & DATA;
		if (flgs == SIZE) size = data;
		else if (flgs == BLOCKED && (data > long_b)) long_b = data;
		else if (flgs == UNBLOCKED && (data > long_u)) long_u = data;
		else if (flgs == FOLLOWER) { /* move on child */
			followers++;
			stack[++stack[0]] = data;
			corridor_eval (data, stack, left, right);
			stack[0]--;
		}
	}
	/* black captures */
	if (long_b + long_u + followers == 0) {
		if (parent != 0) {
			stack[0]--;
			g = corridor_val (corridor_change (stack, FOLLOWER|i,0,0,0));
			stack[0]++;
		}
		else
			g = zero;
		list_insert (left, plus (g, g_int (2 * size)));
	}
	/* white connects to parent*/
	if (parent != 0) {
		l1 = list_copy (inv);
		l2 = list_copy (invasion [parent]);
		list_delete (l1, SIZE | size);
		t = corridor_size (parent);
		if (! (flags & ASSUME2_FLAG)) {
			list_delete (l2, SIZE | t);
			list_insert (l2, SIZE | (t + size + 1));
		}
		list_delete (l2, FOLLOWER | i);
		t = corridor_get (l = list_multi_union (l1, l2));
		if (stack[0] > 2) {
			stack[0] -= 2;
			t = corridor_change (stack, FOLLOWER|stack[stack[0]+1],0,FOLLOWER|t,0);
			stack[0] += 2;
		}
		list_insert (right, corridor_val (t));
		list_free (l);
		list_free (l1);
		list_free (l2);
	} else {
		g = zero;
		for (l = list_start (inv); l != NIL; l = list_rest (l)) {
			n = list_pick (l);
			flgs = FLAGS & n;
			data = DATA & n;
			if (flgs == BLOCKED) g = plus (g, blocked (data));
			else if (flgs == UNBLOCKED) g = plus (g, unblocked (data));
			else if (flgs == FOLLOWER) g = plus (g, corridor_val (data));
		}
		list_insert (right, g);
	}
	list_free (inv);
	if (long_b > 0) {
		/* black blocks longest blocked */
		list_insert (left, plus (g_int (long_b - 1), corridor_val (corridor_change (stack,long_b|BLOCKED,0,0,0))));
		/* white invades longest blocked */
		list_insert (right, corridor_val (corridor_change (stack, long_b|BLOCKED, size|SIZE,long_b==1? 0 : BLOCKED|(long_b-1),
						     (size+(flags & ASSUME2_FLAG ? 0 : 1))|SIZE)));
	}
	if (long_u == 1) myerror ("Corridor_eval: Unblocked corridor length 1");
	else if (long_u > 1) {
		/* black blocks longest unblocked */
		list_insert (left, plus (blocked (long_u-1), corridor_val (corridor_change (stack,UNBLOCKED|long_u,0,0,0))));
		if (! (flags & ASSUME_FLAG)) /* This move should be dominated ?? */
			list_insert (left, corridor_val (corridor_change (stack, UNBLOCKED|long_u, 0, BLOCKED|(long_u-1), 0)));
		/* white invades longest unblocked */
		if ((flags & ASSUME_FLAG) || (long_u == 2)) {
			/* Here we assume move reverses */
			g = plus (corridor_val (corridor_change (stack, UNBLOCKED|long_u,0,0,0)), blocked (long_u-2));
			for (l = list_start (right_options (g)); l!=NIL; l=list_rest (l))
				list_insert (right, list_pick (l));
		} else
			list_insert (right, corridor_val (corridor_change (stack, UNBLOCKED|long_u, 0, UNBLOCKED|(long_u-1), 0)));
	}
	return;
}

game_type corridor_val (int i) {
	list_type left, right;
	int stack[MAX_DEPTH];
	game_type g;

	if (hash_test (CORRIDOR_VALUE_KEY, list_1 (i))) return hash_get_last();
	stack[0] = 1;
	stack[1] = i;
	left = list_make();
	right = list_make();
	corridor_eval (i, stack, left, right);
	g = make (left, right);
	hash_put (CORRIDOR_VALUE_KEY, list_copy_1 (i), g);
	return g;
}

void corridor_Print (int i) {
	int n, data, flgs;
	list_type l;
	printf ("(");
	for (l = list_start (invasion[i]); l != NIL; l = list_rest (l)) {
		n = list_pick (l);
		flgs = n & FLAGS;
		data = n & DATA;
		if (flgs == SIZE) printf ("%d", data);
		else if (flgs == BLOCKED) printf ("b%d", data);
		else if (flgs == UNBLOCKED) printf ("u%d", data);
		else if (flgs == FOLLOWER) corridor_Print (data);
	}
	printf (")");
}

void corridor_printf (int i) {
	corridor_Print (i);
	printf ("\n");
}

int corridor_stat (void) {
	unsigned long i,n=0;
	
	for (i = 1; i <= invasions; i++)
		n = n + list_length (invasion[i]) + 1;
	printf ("corridors: %lu\n", n);
	return n;
}
