/* ====================================================================== *\
   DOMSTR.H -- Evaluate Domineering strings.

   Copyright (C) Jeff Erickson, 1994.  All rights reserved.
   Permission is given to copy this file as part of David Wolfe's
   games program, provided nobody makes any money in the process.
   ======================================================================
   This module calculates values of domineering positions whose
   associated graphs are simple paths, or "strings".  These are
   equivalent to Snort positions in which every node already has a
   color, and the underlying graph is a simple path.

   Evaluating these special positions is considerably easier than
   evaluating generic Domineering positions, for two reasons.  First,
   positions are represented by a string of digits, where each digit
   represents the number of available moves in a segment of the board.
   For example, the position
	   		*
			*
			**
   is denoted "21", and the position
			*
			****
			   *
			   ***
			     ***
			       *
			       *
			       **
			        *
				*
				***
   is denoted "1322123132", as opposed to a 10x11 (22 byte) bitmap
   required by the generic Domineering code.  Different positions
   might be encoded by the same string, but they will all have the
   same value.  For example, "1322123132" also represents the position
	   		***
		     *  * ***
		     ****   *
		     	    *
		           **
			   * 
			   * 
			 ***
   which has the same value (1/32) as the position above.

   Second, positions can be evaluated easily using dynamic programming
   instead of the usual recursive evaluation.  Moreover, all the good
   moves are at the corners.  This implies that if the string has n
   segments, each player has at most n legal moves, and the game will
   be over (except for integer components) after n-1 turns.  Thus,
   evaluating an n-segment string requires only O(n^2) steps,
   regarless of the actual length of each segment.  (In the code, the
   dynamic programming is done implicitly through memoization.)
   
\* ====================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "operations.h"
#include "output.h"
#include "domstr.h"


/* ---------------------------------------------------------------------- *\
|  void DS_print(DS_type pos)
|
|  print a domstr position
\* ---------------------------------------------------------------------- */
void DS_print(DS_type pos)
{
    int i;

    putchar('[');
    if (pos->sign < 0)
	putchar('-');
    for (i=1; i<=pos->length; i++)
    {
	if (pos->board[i] </*=*/ 0)
	    putchar(' ');
	else if (pos->board[i] < 10)
	    putchar(pos->board[i] + '0');
	else
	    putchar(pos->board[i] - 10 + 'A');
    }
    putchar(']');
}

/* ---------------------------------------------------------------------- *\
|  void DS_trace(DS_type pos, game_type g, char *msg, int level)
|
|  Trace output.  Print a position, its value (if known), and an
|  optional message.  Do nothing unless TRACE_FLAG is set.  Inset the
|  output 'level' times. 
\* ---------------------------------------------------------------------- */
void DS_trace(DS_type pos, game_type g, char *msg, int level)
{
    int i;

    if (!(flags&TRACE_FLAG))
	return;
    if ((tlevel != 0) && (level > tlevel))
	return;

    for (i=0; i<level; i++)
	putchar('|'); 
    putchar(' ');

    DS_print(pos);

    if (g != NO_GAME)
    {
	printf(" = ");
	game_print(g);
    }

    if (msg)
	printf("        <%s>", msg);

    putchar('\n');
}

/* ---------------------------------------------------------------------- *\
|  void DS_trace2(DS_type pos1, DS_type pos2, char join, game_type g,
|		  int level)  
|
|  Trace output.  Print the sum or difference of two positions and its
|  value, if known.  Do nothing unless TRACE_FLAG is set.  Inset the
|  output 'level' times.  
\* ---------------------------------------------------------------------- */
void DS_trace2(DS_type pos1, DS_type pos2,
	       char join, game_type g, int level)
{
    int i;

    if (!(flags&TRACE_FLAG))
	return;
    if ((tlevel != 0) && (level > tlevel))
	return;

    for (i=0; i<level; i++)
	putchar('|'); 
    putchar(' ');

    DS_print(pos1);
    printf(" %c ", join);
    DS_print(pos2);

    if (g != NO_GAME)
    {
	printf(" = ");
	game_print(g);
    }

    putchar('\n');
}


/* ---------------------------------------------------------------------- *\
|  void DS_simplify(DS_type pos)
|
|  Simplify a position (in place).
\* ---------------------------------------------------------------------- */
void DS_simplify(DS_type pos)
{
    int i,j;

    for (i=1; (i<=pos->length) && (pos->board[i] == 0); i++)
	pos->sign *= -1;

    for (j=1; (i<=pos->length) && (pos->board[i] > 0); i++, j++)
	pos->board[j] = pos->board[i];

    pos->board[j] = 0;
    pos->length = j-1;
}

/* ---------------------------------------------------------------------- *\
|  DS_type DS_new(int length)
|
|  Get a new domstr position with <length> slots
\* ---------------------------------------------------------------------- */
DS_type DS_new(int length)
{
    DS_type ds;
    int i;

    ds = (DS_type) malloc(sizeof(struct DS));
    ds->length = length;
    ds->sign = 1;
    ds->board = (char *) malloc((length+2) * sizeof(char));
    for (i=0; i<length+2; i++)
	ds->board[i] = 0;
    return ds;
}

/* ---------------------------------------------------------------------- *\
|  void DS_free(DS_type pos)
|
|  Free a domstr position
\* ---------------------------------------------------------------------- */
#define DS_free(pos) if(pos){free((pos)->board);free(pos);}

/* ---------------------------------------------------------------------- *\
|  lisDS_type DS_to_list(DS_type pos)
|
|  Convert a DS_type to a list.  [Wolfe should have called them "sets"
|  instead of "lists", since insertion order doesn't matter!]  Ignores
|  the sign of the position. 
\* ---------------------------------------------------------------------- */
list_type DS_to_list(DS_type pos)
{
    list_type l;
    int i;

    l = list_make();
    for (i=1; i<pos->length+1; i++)
	list_insert(l, 10*i + (int) pos->board[i]);

#ifdef EBUG
    DS_print(pos);
    printf("  ==>  ");
    list_printf(l);
#endif

    return l;
}


/* ---------------------------------------------------------------------- *\
|  DS_type input_domstr(void)
|
|  Ask the user to input a domstr position and return it.
\* ---------------------------------------------------------------------- */
DS_type input_domstr(void)
{
    DS_type pos;
    char s[MAXBUFF];
    int i,j;
    
    printf("Enter position ([1-9A-Z]*):\n");
    gets(s);

    pos = DS_new(strlen(s));

    for (i=0, j=1; s[i]; i++)
    {
	if (s[i] == '0')
	{
	    printf("No zeros allowed!\n");
	    DS_free(pos);
	    return NULL;
	}
	else if (isdigit(s[i]))
	    pos->board[j++] = s[i] - '0';
	else if (isalpha(s[i]))
	    pos->board[j++] = tolower(s[i])-'a'+10;
	else
	{
	    printf("Unknown character '%c'\n", s[i]);
	    DS_free(pos);
	    return NULL;
	}
    }
    pos->length = j-1;
    return pos;
}


/* ---------------------------------------------------------------------- *\
|  game_type domstr_position(DS_type pos, int level)
|
|  Evaluate a simplified single-component domstr position, and memoize
|  the result. 
|
|  DESCRIBE THE ALGORITHM!!
\* ---------------------------------------------------------------------- */
game_type domstr_position(DS_type pos, int level)
{
    register int i,j;
    DS_type new1, new2;
    list_type left, right, cur;
    game_type g = zero;
    game_type h;
    list_type l;

    /* ---- Check for trivial cases ---- */

    if (pos->length == 0)
	return zero;
    
    if (pos->length == 1)
    {
	i = pos->board[1];	/* value = ceiling(length/2) * sign */
	i = i/2 + (i%2);
	i *= pos->sign;

	g = num(Q_int(i));
	return g;
    }

    /* ---- Check for earlier value ---- */
    
    l = DS_to_list(pos);
    if (hash_test(DOMSTR_KEY, l))
    {
	list_free(l);
	g = hash_get_last();

	if (pos->sign < 0)	/* all memoized positions have sign>0 */
	    g = negative(g);

	DS_trace(pos, g, "seen it", level);
	return g;
    }

    /* ---- Recusively evaluate moves ---- */

    DS_trace(pos, NO_GAME, NULL, level);

    new1 = DS_new(pos->length);
    new2 = DS_new(pos->length);

    left = list_make();			/* left options */
    right = list_make();		/* right options */

    for (i=1; i<pos->length+1; i++)
    {
	if (pos->board[i] == 1)		/* one move in this segment */
	{
	    new1->sign = 1;
	    for (j=1; j<i; j++)
		new1->board[j] = pos->board[j];
	    new1->length = i-1;
	    new1->board[i] = 0;
	    if (i > 1)
		new1->board[i-1]--;
	    DS_simplify(new1);

	    new2->sign = 1;
	    for (j=i+1; j<pos->length+1; j++)
		new2->board[j-i] = pos->board[j];
	    new2->length = pos->length - i;
	    new2->board[new2->length+1] = 0;
	    if (new2->length > 0)
		new2->board[1]--;
	    DS_simplify(new2);

	    if (i%2)
	    {
		DS_trace2(new1, new2, '-', NO_GAME, level+1);
		h = minus(domstr_position(new1, level+2),
			  domstr_position(new2, level+2));
		list_insert(left, h);
		DS_trace2(new1, new2, '-', h, level+1);
	    }
	    else
	    {
		DS_trace2(new1, new2, '+', NO_GAME, level+1);
		h = plus(domstr_position(new1, level+2),
			 domstr_position(new2, level+2));
		list_insert(right, h);
		DS_trace2(new1, new2, '+', h, level+1);
	    }
	}
	else if (pos->board[i] > 0)	/* two moves in this segment */
	{
	    if (i<pos->length+1)
	    {
		new1->sign = 1;
		for (j=1; j<=i; j++)
		    new1->board[j] = pos->board[j];
		new1->length = i;
		new1->board[i] -= 2;
		new1->board[i+1] = 0;
		DS_simplify(new1);
		
		new2->sign = 1;
		for (j=i+1; j<pos->length+1; j++)
		    new2->board[j-i] = pos->board[j];
		new2->length = pos->length - i;
		new2->board[new2->length+1] = 0;
		new2->board[1]--;
		DS_simplify(new2);

		if (i%2)
		{
		    DS_trace2(new1, new2, '-', NO_GAME, level+1);
		    h = minus(domstr_position(new1, level+2),
			      domstr_position(new2, level+2));
		    list_insert(left, h);
		    DS_trace2(new1, new2, '-', h, level+1);
		}
		else
		{
		    DS_trace2(new1, new2, '+', NO_GAME, level+1);
		    h = plus(domstr_position(new1, level+2),
			     domstr_position(new2, level+2));
		    list_insert(right, h);
		    DS_trace2(new1, new2, '+', h, level+1);
		}
	    }

	    if (i > 1)
	    {
		new1->sign = 1;
		for (j=1; j<i; j++)
		    new1->board[j] = pos->board[j];
		new1->length = i-1;
		new1->board[i] = 0;
		new1->board[i-1]--;
		DS_simplify(new1);

		new2->sign = 1;
		for (j=i; j<pos->length+1; j++)
		    new2->board[j-i+1] = pos->board[j];
		new2->length = pos->length - i + 1;
		new2->board[new2->length+1] = 0;
		new2->board[1] -= 2;
		DS_simplify(new2);

		if (i%2)
		{
		    DS_trace2(new1, new2, '+', NO_GAME, level+1);
		    h =	plus(domstr_position(new1, level+2),
			     domstr_position(new2, level+2));
		    list_insert(left, h);
		    DS_trace2(new1, new2, '+', h, level+1);
		}
		else
		{
		    DS_trace2(new1, new2, '-', NO_GAME, level+1);
		    h = minus(domstr_position(new1, level+2),
			      domstr_position(new2, level+2));
		    list_insert(right, h);
		    DS_trace2(new1, new2, '-', h, level+1);
		}
	    }
	}
    }

    g = make(left,right);
    hash_put(DOMSTR_KEY, l, g);	/* memoize */

    if (pos->sign < 0)		/* all memoized positions have sign>0 */
	g = negative(g);

    DS_trace(pos, g, NULL, level);

    DS_free(new1);
    DS_free(new2);
    return g;
}


/* ---------------------------------------------------------------------- *\
|  game_type domstr(list_type args)
|
|  Get a Domstr and Frogs position and evaluate it.
\* ---------------------------------------------------------------------- */
game_type domstr(list_type args)
{
    static char varname[MAXBUFF];
    static char base = 'a';

    game_type g,h;
    DS_type pos;
    int length = list_length(args);

    /* ---- evaluate a single position ---- */
    if (length == 0)
    {
	pos = input_domstr();
	if (pos == NULL)
	    return NO_GAME;

	g = domstr_position(pos, 0);
	DS_free(pos);
	return g;
    }

    /* ---- evaluate a series of positions ---- */
    else if (length == 4)
    {
	int prefix, seed, suffix, iterations;
	int i, j;
	int prelength, postlength;
	char prestuff[10], stuff[10], poststuff[10];
	static char varname[MAXBUFF];

	prefix     = list_nth(args, 1);
	seed	   = list_nth(args, 2);
	suffix	   = list_nth(args, 3);
	iterations = list_nth(args, 4);

	if (prefix > 0)
	{
	    sprintf(prestuff, "%d", prefix);
	    for (i=prelength=0; prestuff[i]; i++, prelength++)
		prestuff[i] -= '0';
	}
	else
	    prelength = 0;

	sprintf(stuff, "%d", seed);
	for (i=length=0; stuff[i]; i++, length++)
	    stuff[i] -= '0';

	if (suffix > 0)
	{
	    sprintf(poststuff, "%d", suffix);
	    for (i=postlength=0; poststuff[i]; i++, postlength++)
		poststuff[i] -= '0';
	}
	else
	    postlength = 0;

	pos = DS_new(length*iterations + prelength + postlength);

	for (i=0; i<prelength; i++)
	    pos->board[i+1] = prestuff[i];

	for (i=0; i<iterations; i++)
	{
	    for (j=0; j<length; j++)
		pos->board[prelength + i*length + j + 1] = stuff[j];
	    for (j=0; j<postlength; j++)
		pos->board[prelength + (i+1)*length + j + 1] = poststuff[j];
	    pos->length = prelength + length*(i+1) + postlength;

	    g = domstr_position(pos, 0);

	    sprintf(varname, "%c_%d\0", base, i+1);

	    printf("%s = ", varname);
	    DS_print(pos);
	    printf(" = ");

/*	    if (flags & ISH_FLAG) 	*/
/*	    {				*/
		h = tweak(g);
		if (eq(g, h))
		{
		    game_print(g);
		    printf("\n");
		}
		else
		{
		    game_print(h);
		    printf("-ish\n");
		}
/*	    }				*/
/*	    else			*/
/*	    {				*/
/*		game_print(g);		*/
/*		printf("\n\n");		*/
/*	    }				*/

	    set_var(make_var(varname), g);
	}

	base++;

	DS_free(pos);
	return NO_GAME;
    }
    else
    {
	printf("Don't know how to parse %d arguments.\n", length);
	return NO_GAME;
    }
}
