/* ====================================================================== *\
   WARI.C -- Evaluate Wari positions.

   Copyright (C) Jeff Erickson, 1994.  All rights reserved.
   Permission is given to copy this file as part of David Wolfe's
   games program, provided nobody makes any money in the process.
   ======================================================================
   This module calculates values of Wari positions.  Wari derives
   from a millenia-old African game of the same name, also called
   oware, but the combinatorial formulation here was invented by the
   author.

   The setup is similar to Conway's Sowing.  There is a line of pots,
   each containing some nuimber of stones.  A legal move consists of
   picking up all the stones in one pot, and sowing them in successive
   pots to the right or left.  Unlike Sowing or Mancala, one stone
   must be left in the original pot; consequently, one cannot move
   from a pot containing only one stone.  If the last stone is dropped
   in a pot with exactly one (*) other stone, then BOTH stones are
   removed from the game, or "captured".  Multiple captures are
   possible -- after any capture, the previous pot is examined, and
   if it contains exactly two stones, they are captured.

   For example, moving right from the first pot...

   		41111 -> 12221 -> 10001
		41211 -> 12321 -> 12301
		411   -> 1221
		
   The game ends when no legal moves remain (every pot contains fewer
   then two stones).  The first player who is unable to move
   loses. (**)

   (*) In the traditional African game, piles of 2 or 3, or sometimes
       2 or 4 stones are captured, instead of just two.  Capturing
       groups of 2 stones is necessary to keep the game from being
       loopy.  Allowing ONLY those captures was really an arbitrary
       decision, made in the interest of keeping the game simple.  It
       would be easy to change the implemented rules.

    (**) In the traditional game, the winner is the player who has
         captured the larger number of stones when the game ends.
	 It should be relatively easy to have the terminal positions
	 take on integer values based on the player's scores.
\* ====================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "operations.h"
#include "output.h"
#include "wari.h"


/* ---------------------------------------------------------------------- *\
|  wari_type wari_new(int length)
|
|  Get a new wari position with <length> slots
\* ---------------------------------------------------------------------- */
wari_type wari_new(int length)
{
    wari_type wari;
    int i;

    wari = (wari_type) malloc(sizeof(struct _wari_));
    wari->length = length;
    wari->score = 0;
    wari->board = (char *) malloc((length+2) * sizeof(char));
    for (i=0; i<length+2; i++)
	wari->board[i] = 0;
    return wari;
}

/* ---------------------------------------------------------------------- *\
|  void wari_free(wari_type pos)
|
|  Free a wari position
\* ---------------------------------------------------------------------- */
#define wari_free(pos) if(pos){free((pos)->board);free(pos);}


/* ---------------------------------------------------------------------- *\
|  void wari_print(wari_type pos)
|
|  print a wari position
\* ---------------------------------------------------------------------- */
void wari_print(wari_type pos)
{
    int i, beans;

    for (i = 0; i < pos->length; i++)
    {
	beans = pos->board[i];
	if (beans < 10)
	    putchar(beans + '0');
	else
	    putchar(beans - 10 + 'a');
    }
    if (pos->score)
	printf(" + %d", pos->score);
}

/* ---------------------------------------------------------------------- *\
|  void wari_trace(wari_type pos, game_type g, int s, char *msg, int level)
|
|  Trace output.  Print a position, its value (if known), and an
|  optional message.  Do nothing unless TRACE_FLAG is set.  Inset the
|  output 'level' times.
\* ---------------------------------------------------------------------- */
void wari_trace(wari_type pos, game_type g, int s, char *msg, int level)
{
    int i;

    if (!(flags&TRACE_FLAG))
    {
	if (!s && !msg && (flags & ISH_FLAG))	/* not the best choice */
	{
	    wari_print(pos);
	    printf(" = 0\n");
	}
	return;
    }
    if ((tlevel != 0) && (level > tlevel))
	return;

    for (i=0; i<level; i++)
	putchar('|'); 
    putchar(' ');

    wari_print(pos);

    if (g != NO_GAME)
    {
	printf(" = ");
	game_print(g);
	printf(" [AW ");
	game_print(atomic_weight(g));
	printf("]");
    }
    else if (s != -1)
    {
	if (s == 0)
	    printf(" = 0");
	else if (s == 1)
	    printf(" = *");
	else
	    printf(" = *%d", s);
    }

    if (msg)
	printf("        <%s>", msg);

    putchar('\n');
}


/* ---------------------------------------------------------------------- *\
|  void wari_simplify(wari_type pos)
|
|  Simplify a wari position (in place).
\* ---------------------------------------------------------------------- */
void wari_simplify(wari_type pos)
{
    int i, j;

    for (i=0; i<pos->length; i++)
	if (pos->board[i])
	    break;

    if (i == pos->length)	/* all zeros! */
    {
	pos->length = 0;
	return;
    }
    else if (i > 0)
    {
	for (j=i; j<pos->length; j++)
	    pos->board[j-i] = pos->board[j];
	pos->length -= i;
    }

    while (pos->board[pos->length-1] == 0)
	pos->length--;
}

/* ---------------------------------------------------------------------- *\
|  wari_type input_wari(void)
|
|  Ask the user to input a wari position and return it.
\* ---------------------------------------------------------------------- */
wari_type input_wari(void)
{
    wari_type pos;
    char s[MAXBUFF];
    int i,j;
    char space=1;
    
    printf("Enter position (list of hexidecimal digits):\n");
    gets(s);

    pos = wari_new(strlen(s));

    for (i=j=0; s[i]; i++)
    {
	if (isspace(s[i]))
	    continue;
	else if (isdigit(s[i]))
	    pos->board[j++] = s[i] - '0';
	else if (isalpha(s[i]) && (tolower(s[i]) <= 'f'))
	    pos->board[j++] = tolower(s[i]) - 'a' + 10;
	else if (tolower(s[i]) == 'x')
	    pos->board[j++] = -1;
	else
	{
	    printf("Unknown character '%c'\n", s[i]);
	    wari_free(pos);
	    return NULL;
	}
    }
    pos->length = j;
    wari_simplify(pos);
    return pos;
}


/* ---------------------------------------------------------------------- *\
|  list_type wari_to_list(wari_type pos)
|
|  Convert a wari_type to a list.  [Wolfe should have called them "sets"
|  instead of "lists", since insertion order doesn't matter!]  Saves
|  space by combining small integers into larger ones.  What this
|  program really needs is the ability to has STRINGS.
\* ---------------------------------------------------------------------- */
list_type wari_to_list(wari_type pos)
{
    list_type l;
    int i, base, digit;

    l = list_make();
    list_insert(l, pos->length);
    list_insert(l, pos->score * 100);

    base = 2;
    for (i=0; i<pos->length; i++)
	if (base < pos->board[i] + 2)
	    base = pos->board[i] + 2;
    list_insert(l, base*10000);

    digit = 0;
    for (i = 0; i < pos->length; i++)
    {
	if (digit*base > (1L<<20))
	{
	    list_insert(l, (i<<20) + digit);
	    digit = 0;
	}
	digit *= base;
	
	digit += pos->board[i] + 1;
    }
    if (digit != 0)
	list_insert(l, (i<<20) + digit);

#ifdef EBUG
    wari_print(pos);
    printf("  ==>  ");
    list_printf(l);
#endif

    return l;
}

/* ---------------------------------------------------------------------- *\
|  list_type symmetric_wari_to_list(wari_type pos)
|
|  Convert a wari_type to a list.  Compares the position with its
|  mirror image and build the list from the lexicographically smaller
|  of the pair.  Saves space by combining small integers into larger
|  ones.  What this program really needs is a good five-cent cigar.
\* ---------------------------------------------------------------------- */
list_type symmetric_wari_to_list(wari_type pos)
{
    list_type l;
    int i, base, sign;
    int beans, snaeb, digit;

    l = list_make();
    list_insert(l, pos->length);
    list_insert(l, pos->score * 100);

    for (sign=i=0; !sign && (1+i <= pos->length-i); i++)
    {
	beans = pos->board[i];
	if (beans < 0)
	    beans = 10000;

	snaeb = pos->board[pos->length - i - 1];
	if (snaeb < 0)
	    snaeb = 10000;

	sign = snaeb - beans;
    }

    base = 2;
    for (i=0; i<pos->length; i++)
	if (base < pos->board[i] + 2)
	    base = pos->board[i] + 2;
    list_insert(l, base*10000);

    if (sign < 0)
    {
	digit = 0;
	for (i = 0; i < pos->length; i++)
	{
	    if (digit*base > (1L<<20))
	    {
		list_insert(l, (i<<20) + digit);
		digit = 0;
	    }
	    digit *= base;
	    digit += pos->board[pos->length - i - 1] + 1;
	}
	if (digit != 0)
	    list_insert(l, (i<<20) + digit);
    }
    else
    {
	digit = 0;
	for (i = 0; i < pos->length; i++)
	{
	    if (digit*base > (1L<<20))
	    {
		list_insert(l, (i<<20) + digit);
		digit = 0;
	    }
	    digit *= base;
	    digit += pos->board[i] + 1;
	}
	if (digit != 0)
	    list_insert(l, (i<<20) + digit);
    }

#ifdef EBUG
    putchar(sign<0 ? '-' : '+');
    wari_print(pos);
    printf(" (%d to %d) ==>  ", 1, pos->length);
    list_printf(l);
#endif

    return l;
}


/* ---------------------------------------------------------------------- *\
|  int deuce(wari_type pos)
|
|  Evaluate a wari position where every pot contains 2 or fewer
|  stones, and return the nim value.  NOTE THAT THIS FUNCTION RETURNS
|  AN INTEGER, NOT A GAME_TYPE!!
|
|  Returns -1 if the position is too big to compute directly.
\* ---------------------------------------------------------------------- */
int wari_deuce(wari_type pos)
{
    int i, count, result;
    
    static int nimval[] =	 	/* Nim values of strings of 2's */
    {
	0, 1, 2, 0, 1, 2, 3, 1, 2, 3, 4, 0, 3, 4, 2, 1,
	3, 2, 1, 0, 2, 1, 4, 5, 1, 4, 5, 1, 2, 0, 1, 2
    };

    count = 0;
    result = 0;
    for (i=0; i<pos->length; i++)
    {
	if (pos->board[i] == 2)
	{
	    count++;
	    if (count > 31)
		return -1;
	}
	else if (count)
	{
	    result ^= nimval[count];
	    count = 0;
	}
    }
    if (count)
	result ^= nimval[count];

    return result;
}


/* ---------------------------------------------------------------------- *\
|  int iwari_position(wari_type pos, int level)
|
|  Evaluate a impartial wari position, and memoize the result.
|  NOTE THAT THIS FUNCTION RETURNS AN INTEGER, NOT A GAME_TYPE!!
\* ---------------------------------------------------------------------- */
int iwari_position(wari_type pos, int level)
{
    register int i, j, start, beans;
    int ace, deuce;
    list_type options;
    int g;		/* really a game */
    list_type l;
    wari_type new;

    /* ---- Check for trivial cases ---- */

    if (pos->length < 1)
	return 0;

    /* ---- Check for earlier value ---- */
    
    l = symmetric_wari_to_list(pos);
    if (hash_test(IWARI_KEY, l))
    {
	list_free(l);
	g = hash_get_last();
	wari_trace(pos, NO_GAME, g, "seen it", level);
	return g;
    }

    /* ---- Look for deuces ---- */

    ace = deuce = 1;
    for (i=0; deuce && (i<pos->length); i++)
    {
	deuce &= (pos->board[i]<3);
	ace &= (pos->board[i]<2);
    }

    if (ace)
    {
	wari_trace(pos, NO_GAME, 0, "trivial", level);
	hash_put(IWARI_KEY, l, 0);		/* memoize */
	return 0;
    }
    if (deuce)
    {
	g = wari_deuce(pos);
	if (g != -1)
	{
	    wari_trace(pos, NO_GAME, g, "deuce", level);
	    hash_put(IWARI_KEY, l, g);		/* memoize */
	    return g;
	}
    }

    /* ---- Recusively evaluate moves ---- */

    wari_trace(pos, NO_GAME, -1, NULL, level);

    options = list_make();

    for (i=0; i<pos->length; i++)
    {
	beans = pos->board[i];
	if (beans <= 1)
	    continue;

	/* ---- right move ---- */

	if (i >= beans)
	{
	    new = wari_new(pos->length);
	    for (j=0; j<pos->length; j++)
		new->board[j] = pos->board[j];
	    start = i;
	}
	else
	{
	    new = wari_new(pos->length - i + beans - 1);
	    for (j=0; j<pos->length; j++)
		new->board[j-i+beans-1] = pos->board[j];
	    start = beans-1;
	}

	new->board[start] = 0;
	for (j=0; j<beans; j++)
	    new->board[start - j]++;

	for (j=start-beans+1; (new->board[j]==2 || new->board[j]==3); j++)
	    new->board[j] = 0;

	wari_simplify(new);
	list_insert(options, iwari_position(new, level+1));
	wari_free(new);

	/* ---- left move ---- */

	if (pos->length > i+ beans)
	    new = wari_new(pos->length);
	else
	    new = wari_new(i + beans);

	for (j=0; j<pos->length; j++)
	    new->board[j] = pos->board[j];
	start = i;

	new->board[start] = 0;
	for (j=0; j<beans; j++)
	    new->board[start + j]++;

	for (j=start+beans-1; (new->board[j]==2 || new->board[j]==3); j--)
	    new->board[j] = 0;

	wari_simplify(new);
	list_insert(options, iwari_position(new, level+1));
	wari_free(new);
    }

    g = mex(options);
    list_free(options);

    hash_put(IWARI_KEY, l, g);		/* memoize */

    wari_trace(pos, NO_GAME, g, NULL, level);
    return g;
}


/* ---------------------------------------------------------------------- *\
|  game_type wari_position(wari_type pos, int level)
|
|  Evaluate a partizan wari position, and memoize the result.
\* ---------------------------------------------------------------------- */
game_type wari_position(wari_type pos, int level)
{
    register int i, j, start, beans;
    int score;
    int ace, deuce, count;
    list_type left, right, options;
    game_type g;
    wari_type new;
    list_type l;

    /* ---- Check for trivial cases ---- */

    if (pos->length < 1)
	return zero;

    score = pos->score;
    pos->score = 0;

    l = wari_to_list(pos);  	 	/* check for earlier value */
    if (hash_test(WARI_KEY, l))
    {
	list_free(l);
	g = hash_get_last();
	wari_trace(pos, g, -1, "seen it", level);
    }
    else
    {
#if 1
	/* ---- Look for deuces ---- */
	
	ace = deuce = 1;
	for (i=0; deuce && (i<pos->length); i++)
	{
	    ace &= (pos->board[i]<2);
	    deuce &= (pos->board[i]<3);
	}

	if (ace)
	{
	    wari_trace(pos, zero, -1, "trivial", level);
	    hash_put(IWARI_KEY, l, zero); /* memoize */
	    return zero;
	}
	if (deuce)
	{
	    count = wari_deuce(pos);
	    if (count != -1)
	    {
		g = star(count);
		wari_trace(pos, g, -1, "deuce", level);
		hash_put(IWARI_KEY, l, g); /* memoize */
		return g;
	    }
	}
#endif
	wari_trace(pos, NO_GAME, -1, NULL, level);

	left = list_make();		/* left options */
	right = list_make();		/* right options */
	
	for (i=0; i<pos->length; i++)
	{
	    beans = pos->board[i];
	    if (beans <= 1)
		continue;

	    /* ---- right move ---- */

	    if (i >= beans)
	    {
		new = wari_new(pos->length);
		for (j=0; j<pos->length; j++)
		    new->board[j] = pos->board[j];
		start = i;
	    }
	    else
	    {
		new = wari_new(pos->length - i + beans - 1);
		for (j=0; j<pos->length; j++)
		    new->board[j-i+beans-1] = pos->board[j];
		start = beans-1;
	    }

	    new->board[start] = 0;
	    for (j=0; j<beans; j++)
		new->board[start - j]++;

	    for (j=start-beans+1; (new->board[j]==2 || new->board[j]==3); j++)
	    {
		new->board[j] = 0;
#if 0
		new->score -= 2;
#endif
	    }

	    wari_simplify(new);
	    list_insert(right, wari_position(new, level+1));
	    wari_free(new);

	    /* ---- left move ---- */

	    if (pos->length > i+ beans)
		new = wari_new(pos->length);
	    else
		new = wari_new(i + beans);

	    for (j=0; j<pos->length; j++)
		new->board[j] = pos->board[j];
	    start = i;

	    new->board[start] = 0;
	    for (j=0; j<beans; j++)
		new->board[start + j]++;

	    for (j=start+beans-1; (new->board[j]==2 || new->board[j]==3); j--)
	    {
		new->board[j] = 0;
#if 0
		new->score += 2;
#endif
	    }

	    wari_simplify(new);
	    list_insert(left, wari_position(new, level+1));
	    wari_free(new);
	}
	
	g = make(left,right);
	
	hash_put(WARI_KEY, l, g);		/* memoize */
	
	wari_trace(pos, g, -1, NULL, level);
    }

    if (score)
    {
	g = plus(g, num(Q_int(score)));
	pos->score = score;
	wari_trace(pos, g, -1, "+ score", level);
    }

    return g;
}


/* ---------------------------------------------------------------------- *\
|  game_type wari(int impartial, list_type args)
|
|  Get a Wari position and evaluate it.
\* ---------------------------------------------------------------------- */
game_type wari(int impartial, list_type args)
{
    wari_type pos;
    game_type g;

    pos = input_wari();
    if (pos == NULL)
	return NO_GAME;
    
    if (impartial)
	g = star(iwari_position(pos, 0));
    else
	g = wari_position(pos, 0);

    wari_free(pos);
    return g;
}
